@@ -4,7 +4,7 @@ from tkinter import scrolledtext
 from tkinter import font
 from functools import partial
 from gooeypie.error import GooeyPieError
-from .containers import *   # Used for the Listbox & Radiogroup widgets
+from .containers import *  # Used for the Listbox & Radiogroup widgets
 
 import platform
 
@@ -23,6 +23,7 @@ except ImportError:
 
 class GooeyPieEvent:
     """Event objects are passed to callback functions"""
+
     def __init__(self, event_name, gooeypie_widget, tk_event=None, menu=None):
         """Creates a GooeyPie event object, passed to all callback functions"""
         self.event_name = event_name
@@ -44,8 +45,8 @@ class GooeyPieEvent:
             else:
                 self.key = {
                     'key_code': tk_event.char,  # single character representing the key
-                    'name': tk_event.keysym,    # a string representing the key pressed
-                    'code': tk_event.keycode    # numerical keycode
+                    'name': tk_event.keysym,  # a string representing the key pressed
+                    'code': tk_event.keycode  # numerical keycode
                 }
         else:
             self.mouse = None
@@ -54,7 +55,7 @@ class GooeyPieEvent:
         # The menu path if a menu event was triggered
         self.menu = menu
 
-    def __repr__(self):
+    def __str__(self):
         return str({
             'event_name': self.event_name,
             'widget': self.widget,
@@ -63,6 +64,9 @@ class GooeyPieEvent:
             'menu': self.menu
         })
 
+    def __repr__(self):
+        return self.__str__()
+
 
 class GooeyPieWidget:
     """Base class for other GooeyPie widget classes, mostly for event handling"""
@@ -126,6 +130,16 @@ class GooeyPieWidget:
         3 arguments to the callback"""
         self._event(event_name)
 
+    def _textbox_change_event(self, event_name, key_code):
+        """To implement the change event on ScrolledText, a sentinel variable is associated with the contents
+        of the text widget. This value is updated
+        """
+        if self.text != self._sentinel.get():
+            # Only process the event if the key release was a change in content, thus ignoring modifier keys etc
+            self._sentinel.set(self.text)
+            self.text = self._sentinel.get()  # just in case
+            self._event(event_name)
+
     def add_event_listener(self, event_name, callback):
         """Registers callback to respond to certain events"""
 
@@ -140,7 +154,7 @@ class GooeyPieWidget:
 
         # Check that the event function specified accepts a single argument
         if callback.__code__.co_argcount != 1:
-            raise GooeyPieError(f"Your event function '{callback.__name__}' must accept a single argument")
+            raise GooeyPieError(f"The event function '{callback.__name__}' must accept a single argument")
 
         # Hyperlinks have default events for mouse_down (activating the link)
         # and mouse_over (showing a hand icon)
@@ -153,8 +167,11 @@ class GooeyPieWidget:
 
         if event_name in self._tk_event_mappings:
             if isinstance(self, Listbox):
-                # Bind the event to the listbox part of the ScrolledListbox
+                # Bind the event to the listbox part of the Listbox widget
                 self._listbox.bind(self._tk_event_mappings[event_name], partial(self._event, event_name))
+            elif isinstance(self, Table):
+                # Bind the event to the treeview part of the Table widget
+                self._treeview.bind(self._tk_event_mappings[event_name], partial(self._event, event_name))
             else:
                 self.bind(self._tk_event_mappings[event_name], partial(self._event, event_name))
 
@@ -186,22 +203,22 @@ class GooeyPieWidget:
                 self._observer = self._value.trace('w', partial(self._text_change_event, event_name))
 
             if isinstance(self, Textbox):
-                # TODO: change event for the textbox is complicated - will need to add a 'sentinel' to the Textbox widget
-                # http://webcache.googleusercontent.com/search?q=cache:KpbCmAzvn_cJ:code.activestate.com/recipes/464635-call-a-callback-when-a-tkintertext-is-modified/+&cd=2&hl=en&ct=clnk&gl=au
-                self.bind('<<Modified>>', partial(self._event, event_name))
+                self.bind('<KeyRelease>', partial(self._textbox_change_event, event_name))
 
         if event_name == 'press':
-            # press event only on buttons (for now perhaps...)
+            # press event only on buttons
             self.configure(command=partial(self._event, event_name))
 
         if event_name == 'select':
-            # Select event associated at the moment with listboxes and dropdowns
+            # Select event associated with Listboxes, dropdowns and tables
             if isinstance(self, SimpleListbox):
                 self.bind('<<ListboxSelect>>', partial(self._event, event_name))
             elif isinstance(self, Listbox):
                 self._listbox.bind('<<ListboxSelect>>', partial(self._event, event_name))
             elif isinstance(self, Dropdown):
                 self.bind('<<ComboboxSelected>>', partial(self._event, event_name))
+            elif isinstance(self, Table):
+                self._treeview.bind('<<TreeviewSelect>>', partial(self._event, event_name))
 
     def remove_event_listener(self, event_name):
         """
@@ -234,14 +251,13 @@ class GooeyPieWidget:
                 self.unbind('<Return>')
 
             if isinstance(self, Input):
-                # If there is a change event, then
+                # If there is a change event, then delete the trace
                 if self._observer:
                     self._value.trace_vdelete('w', self._observer)
                     self._observer = None
 
             if isinstance(self, Textbox):
-                # TODO: add_event_listener not implemented
-                self.unbind('<<Modified>>')
+                self.unbind('<KeyRelease>')
 
         if event_name == 'press':
             # press event on buttons
@@ -255,6 +271,8 @@ class GooeyPieWidget:
                 self._listbox.unbind('<<ListboxSelect>>')
             elif isinstance(self, Dropdown):
                 self.unbind('<<ComboboxSelected>>')
+            elif isinstance(self, Table):
+                self._treeview.unbind('<<TreeviewSelect>>')
 
     # All widgets can be enabled and disabled
     @property
@@ -276,6 +294,12 @@ class GooeyPieWidget:
             state = 'disabled' if self._disabled else 'normal'
             self._listbox.config(state=state)
 
+        elif isinstance(self, Table):
+            # The treeview is a member of the Table object
+            # Note: events still fire when the table is disabled
+            state = ['disabled'] if self._disabled else ['!disabled']
+            self._treeview.state(state)
+
         elif isinstance(self, RadiogroupBase):
             # Both the container and each radiobutton are disabled
             state = ['disabled'] if self._disabled else ['!disabled']
@@ -283,16 +307,17 @@ class GooeyPieWidget:
             for radio in self.winfo_children():
                 radio.state(state)  # disable each radiobutton
 
+        elif isinstance(self, Hyperlink):
+            # Change the hand cursor to the 'no' cursor for a disabled Hyperlink
+            self.cursor = 'no' if self._disabled else 'hand2'
+            state = ['disabled'] if self._disabled else ['!disabled']
+            self.state(state)
+
         else:
             # most other widgets are ttk widgets disabled with the state() method
             state = ['disabled'] if self._disabled else ['!disabled']
             self.state(state)
 
-    # TESTING: Retrieve all grid settings
-    def get_info(self):
-        return self.grid_info()
-        # Sample output: Grid info: {'in': <guilite.GuiLiteApp object .!guiliteapp>, 'column': 1, 'row': 1, 'columnspan': 1, 'rowspan': 1, 'ipadx': 0, 'ipady': 0, 'padx': (0, 10), 'pady': (0, 10), 'sticky': 'nw'}
-
     @property
     def margin_top(self):
         return self.margins[0]
@@ -354,6 +379,9 @@ class Label(ttk.Label, GooeyPieWidget):
     def __str__(self):
         return f"<Label '{self.text}'>"
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def text(self):
         return self.cget('text')
@@ -450,10 +478,16 @@ class Button(ttk.Button, GooeyPieWidget):
     def __str__(self):
         return f"<Button '{self.text}'>"
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def width(self):
         return self.cget('width')
 
+    def __repr__(self):
+        return self.__str__()
+
     @width.setter
     def width(self, value):
         self.configure(width=value)
@@ -466,10 +500,6 @@ class Button(ttk.Button, GooeyPieWidget):
     def text(self, text):
         self.configure(text=text)
 
-    # TESTING #
-    def info(self):
-        return self.grid_info()
-
 
 class Slider(ttk.Scale, GooeyPieWidget):
     def __init__(self, container, low, high, orientation='horizontal'):
@@ -494,6 +524,9 @@ class Slider(ttk.Scale, GooeyPieWidget):
     def __str__(self):
         return f'<Slider from {self.cget("from")} to {self.cget("to")}>'
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def value(self):
         return self._value.get()
@@ -525,6 +558,9 @@ class StyleLabel(Label):
     def __str__(self):
         return f"<StyleLabel '{self.text}'>"
 
+    def __repr__(self):
+        return self.__str__()
+
     def _get_current_font(self):
         """Returns a dictionary representing the current font"""
         return font.Font(font=self._style.lookup(self._style_id, 'font')).actual()
@@ -700,7 +736,7 @@ class StyleLabel(Label):
         for n, x in enumerate(layout):
             if x == '(':
                 element = ""
-                for y in layout[n+2:]:
+                for y in layout[n + 2:]:
                     if y != ',':
                         element += str(y)
                     else:
@@ -719,13 +755,17 @@ class Hyperlink(StyleLabel):
         self.url = url
         self.colour = 'blue'
         self.underline = 'underline'
-        self.bind('<Enter>', lambda e: self.configure(cursor='hand2'))
+        self.cursor = 'hand2'
+        self.bind('<Enter>', lambda e: self.configure(cursor=self.cursor))
         self.bind('<Button-1>', self._open_link)
         self.configure(takefocus=True)  # Labels don't normally take focus when tabbing
 
     def __str__(self):
         return f"<Hyperlink '{self.text}'>"
 
+    def __repr__(self):
+        return self.__str__()
+
     def _open_link(self, e):
         if not self.disabled:
             import webbrowser
@@ -740,27 +780,28 @@ class Image(Label):
     def __str__(self):
         return f"""<Image '{self.image}'>"""
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def image(self):
         return self._image
 
     @image.setter
-    def image(self, image):
-        image_extension = image[-3:]
+    def image(self, image_path):
+        image_extension = image_path[-3:]
         if not PILLOW and image_extension != 'gif':
             raise ValueError('Only gif images can be used at this time.')
 
-        self._image = image
+        self._image = image_path
 
         if not PILLOW:
-            self._tk_image = tk.PhotoImage(file=image)
+            self._tk_image = tk.PhotoImage(file=image_path)
         else:
-            self._tk_image = ImageTk.PhotoImage(PILImage.open(image))
+            self._tk_image = ImageTk.PhotoImage(PILImage.open(image_path))
 
         self.configure(image=self._tk_image)
 
-    # TODO: Overwrite the text setter to raise an error?
-
 
 class Input(ttk.Entry, GooeyPieWidget):
     def __init__(self, container):
@@ -773,6 +814,9 @@ class Input(ttk.Entry, GooeyPieWidget):
     def __str__(self):
         return f"""<Input widget>"""
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def width(self):
         return self.cget('width')
@@ -827,6 +871,9 @@ class Secret(Input):
     def __str__(self):
         return f"""<Secret widget>"""
 
+    def __repr__(self):
+        return self.__str__()
+
     def unmask(self):
         self.configure(show='')
 
@@ -861,6 +908,9 @@ class SimpleListbox(tk.Listbox, GooeyPieWidget):
     def __str__(self):
         return f'<Listbox {tuple(self.items)}>'
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def height(self):
         """Returns the number of lines in the listbox"""
@@ -868,8 +918,7 @@ class SimpleListbox(tk.Listbox, GooeyPieWidget):
 
     @height.setter
     def height(self, lines):
-        """Sets the *minimum* number of lines in the listbox. The number of visible lines may be different
-        """
+        """Sets the minimum number of lines in the listbox."""
         self.configure(height=lines)
 
     @property
@@ -879,7 +928,7 @@ class SimpleListbox(tk.Listbox, GooeyPieWidget):
 
     @width.setter
     def width(self, chars):
-        """Sets the width of the listbox, in characters. Default is 20."""
+        """Sets the minimum width of the listbox, in characters. Default is 20."""
         self.configure(width=chars)
 
     @property
@@ -1016,7 +1065,7 @@ class SimpleListbox(tk.Listbox, GooeyPieWidget):
                 # Return single value
                 index = self.selected_index
                 removed_item = self.remove_item(index)
-                self.selected_index = index   # Keep the existing selection
+                self.selected_index = index  # Keep the existing selection
                 return removed_item
 
 
@@ -1199,6 +1248,7 @@ class Textbox(scrolledtext.ScrolledText, GooeyPieWidget):
     def __init__(self, container, width=20, height=5):
         GooeyPieWidget.__init__(self, container)
         scrolledtext.ScrolledText.__init__(self, container, width=width, height=height)
+        self._sentinel = tk.StringVar()
 
         self.configure(borderwidth=1, relief='flat', font=font.nametofont('TkDefaultFont'),
                        wrap='word', highlightcolor='systemHighlight', highlightthickness=1)
@@ -1213,14 +1263,14 @@ class Textbox(scrolledtext.ScrolledText, GooeyPieWidget):
         self.bind('<Tab>', self.focus_next_widget)
         self.bind('<Shift-Tab>', self.focus_previous_widget)
         self.bind('<Control-Tab>', self.insert_tab)
-        # self._events['change'] = None
-
-    # TODO: readonly option
-    # https://stackoverflow.com/questions/3842155/is-there-a-way-to-make-the-tkinter-text-widget-read-only
+        self._events['change'] = None
 
     def __str__(self):
         return f"""<Textbox object>"""
 
+    def __repr__(self):
+        return self.__str__()
+
     @staticmethod
     def focus_next_widget(event):
         """Overrides the default behaviour of inserting a tab character in a textbox instead of
@@ -1311,6 +1361,9 @@ class ImageButton(Button):
     def __str__(self):
         return f"""<ImageButton '{self._image}'>"""
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def image_position(self):
         return self.cget('compound')
@@ -1327,11 +1380,14 @@ class Checkbox(ttk.Checkbutton, GooeyPieWidget):
         self._checked = tk.BooleanVar(value=False)
         ttk.Checkbutton.__init__(self, container, text=text, variable=self._checked)
         self.state(['!alternate'])
-        self._events['change'] = None   # Checkboxes support the 'change' event
+        self._events['change'] = None  # Checkboxes support the 'change' event
 
     def __str__(self):
         return f'''<Checkbox '{self.cget("text")}'>'''
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def checked(self):
         return self._checked.get()
@@ -1343,9 +1399,10 @@ class Checkbox(ttk.Checkbutton, GooeyPieWidget):
 
 class RadiogroupBase(GooeyPieWidget):
     """Base class used by Radiogroup and LabelledRadiogroup"""
+
     def __init__(self, container, choices, orient, override_spacing=False):
         GooeyPieWidget.__init__(self, container)
-        self._events['change'] = None   # Radiobuttons support the 'change' event
+        self._events['change'] = None  # Radiobuttons support the 'change' event
         self._selected = tk.StringVar()
 
         # If images are used (to be implemented, need to support passing a list of 2-tuples,
@@ -1384,7 +1441,6 @@ class RadiogroupBase(GooeyPieWidget):
             if isinstance(self, LabelRadiogroup):
                 margins = ['auto'] * 4
                 # For vertically aligned radiogroups, reduce the vertical spacing between items.
-
                 if orient == 'vertical':
                     if pos != length - 1:
                         margins[2] = 0
@@ -1421,6 +1477,7 @@ class RadiogroupBase(GooeyPieWidget):
 
 class Radiogroup(Container, RadiogroupBase):
     """A set of radio buttons"""
+
     def __init__(self, container, choices, orient='vertical'):
         Container.__init__(self, container)
         RadiogroupBase.__init__(self, container, choices, orient, True)
@@ -1428,9 +1485,13 @@ class Radiogroup(Container, RadiogroupBase):
     def __str__(self):
         return f'<Radiogroup {tuple(self.options)}>'
 
+    def __repr__(self):
+        return self.__str__()
+
 
 class LabelRadiogroup(LabelContainer, RadiogroupBase):
     """A set of radio buttons in a label frame"""
+
     def __init__(self, container, title, choices, orient='vertical'):
         LabelContainer.__init__(self, container, title)
         RadiogroupBase.__init__(self, container, choices, orient)
@@ -1438,58 +1499,8 @@ class LabelRadiogroup(LabelContainer, RadiogroupBase):
     def __str__(self):
         return f'<LabelRadiogroup {tuple(self.options)}>'
 
-
-
-class OldRadiogroupBase(GooeyPieWidget):
-    """Base class used by Radiogroup and LabelledRadiogroup"""
-    def __init__(self, container, choices, orient):
-        GooeyPieWidget.__init__(self, container)
-        self._events['change'] = None   # Radiobuttons support the 'change' event
-        self._selected = tk.StringVar()
-
-        # If images are used (to be implemented, need to support passing a list of 2-tuples,
-        # where first item is the image, second is the value returned
-
-        if isinstance(choices, (list, tuple)):
-            side = 'left' if orient == 'horizontal' else 'top'
-            for choice in choices:
-                radiobutton = ttk.Radiobutton(self, text=choice, variable=self._selected, value=choice)
-                radiobutton.pack(expand=True, fill='x', padx=8, pady=8, side=side)
-
-    @property
-    def options(self):
-        return tuple(widget.cget('text') for widget in self.winfo_children())
-
-    @property
-    def selected(self):
-        if self._selected.get():
-            return self._selected.get()
-        else:
-            return None
-
-    @selected.setter
-    def selected(self, value):
-        self._selected.set(value)
-
-
-class OldRadiogroup(ttk.Frame, RadiogroupBase):
-    """A set of radio buttons"""
-    def __init__(self, container, choices, orient='vertical'):
-        ttk.Frame.__init__(self, container)
-        RadiogroupBase.__init__(self, container, choices, orient)
-
-    def __str__(self):
-        return f'<Radiogroup {tuple(self.options)}>'
-
-
-class OldLabelRadiogroup(ttk.LabelFrame, RadiogroupBase):
-    """A set of radio buttons in a label frame"""
-    def __init__(self, container, title, choices, orient='vertical'):
-        ttk.LabelFrame.__init__(self, container, text=title)
-        RadiogroupBase.__init__(self, container, choices, orient)
-
-    def __str__(self):
-        return f'<LabelRadiogroup {tuple(self.options)}>'
+    def __repr__(self):
+        return self.__str__()
 
 
 class Dropdown(ttk.Combobox, GooeyPieWidget):
@@ -1499,14 +1510,12 @@ class Dropdown(ttk.Combobox, GooeyPieWidget):
         self.state(['readonly'])
         self._events['select'] = None
 
-        # DONE TODO 1: change choices to items
-        # DONE TODO 2: need to write getters and setters for choices - at the moment nothing actually changes...
-        # TODO 3: update docs
-        # TODO 4: write a better testing widget
-
     def __str__(self):
         return f'<Dropdown {tuple(self.items)}>'
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def items(self):
         """Returns a COPY of the items in the Dropdown"""
@@ -1577,6 +1586,9 @@ class Number(ttk.Spinbox, GooeyPieWidget):
         high = type(self.cget("increment"))(self.cget("to"))
         return f'<Number widget from {low} to {high}>'
 
+    def __repr__(self):
+        return self.__str__()
+
     @property
     def value(self):
         """Returns the value of the Number widget according to the type of increment
@@ -1664,12 +1676,271 @@ class Number(ttk.Spinbox, GooeyPieWidget):
         self.configure(wrap=bool(state))
 
 
-class Table(ttk.Treeview, GooeyPieWidget):
+class Table(Container, GooeyPieWidget):
     """For displaying tabular data"""
-    def columns(self, cols):
-        """Sets the column names of the table"""
-        pass
+    icon_spacing = '   '
+    sort_ascending_icon = f'{icon_spacing}▲'
+    sort_descending_icon = f'{icon_spacing}▼'
 
-    def add_data(self, data):
+    def __init__(self, container, headings):
+        Container.__init__(self, container)
+
+        # Check that the heading are in a list
+        if type(headings) != list and type(headings) != tuple:
+            raise ValueError(f'Headings must be a list. Argument was: {type(headings)}')
+
+        # Set container to fill cell
+        self.columnconfigure(0, weight=1)
+        self.rowconfigure(0, weight=1)
+
+        # Create and configure treeview
+        self._num_columns = len(headings)
+        column_ids = tuple(range(self._num_columns))  # tuple of form (0, 1, 2, etc)
+        self._treeview = ttk.Treeview(self, columns=column_ids, show='headings', selectmode='browse')
+        for index, heading in enumerate(headings):
+            self._treeview.heading(index, text=heading, command=lambda col_id=index: self._sort_data(col_id))
+
+        # Left align cell contents to center by default
+        for col_id in column_ids:
+            self._treeview.column(col_id, anchor='w')
+
+        # Create vertical scrollbar configure behaviour
+        self._v_scrollbar = ttk.Scrollbar(self, orient='vertical')
+        self._v_scrollbar.config(command=self._treeview.yview)
+        self._treeview.config(yscroll=self._v_scrollbar.set)
+
+        # create horizontal scrollbar and configure behaviour
+        self._h_scrollbar = ttk.Scrollbar(self, orient='horizontal')
+        self._h_scrollbar.config(command=self._treeview.xview)
+        self._treeview.config(xscroll=self._h_scrollbar.set)
+
+        # Add to parent Container
+        self._treeview.grid(row=0, column=0, sticky='nsew')
+        self._v_scrollbar.grid(row=0, column=1, sticky='nsew')
+        self._h_scrollbar.grid(row=1, column=0, sticky='nsew')
+
+        # Default scrollbar settings and bindings to update visibility of scrollbars
+        self._treeview.bind('<Configure>', self._update_scrollbar)  # Update scrollbar visibility when widget changes size
+        self._treeview.bind('<ButtonRelease-1>', self._update_scrollbar)
+
+        GooeyPieWidget.__init__(self, container)
+        self._events['select'] = None
+
+    def __str__(self):
+        # Identified by column names
+        return f"<Table {tuple([self._treeview.heading(col_id)['text'] for col_id in range(self._num_columns)])}>"
+
+    def __repr__(self):
+        return self.__str__()
+
+    @property
+    def height(self):
+        """Returns the number of visible row in the table"""
+        return self._treeview.cget('height')
+
+    @height.setter
+    def height(self, lines):
+        """Sets the*minimum number of visible rows in the table. The number of visible lines may be different"""
+        self._treeview.configure(height=lines)
+
+    def _update_scrollbar(self, _event=None):
+        """Adds/removes the horizontal scrollbar as needed"""
+        horizontal_scrollbar_needed = self._treeview.xview() != (0.0, 1.0)
+        row_span = 1 if horizontal_scrollbar_needed else 2
+
+        self._treeview.grid_remove()
+        self._v_scrollbar.grid_remove()
+        self._h_scrollbar.grid_remove()
+
+        self._treeview.grid(row=0, column=0, rowspan=row_span)
+        self._v_scrollbar.grid(row=0, column=1, rowspan=row_span)
+        if horizontal_scrollbar_needed:
+            self._h_scrollbar.grid()
+
+    def _sort_data(self, column_id):
+        """When the column heading is clicked on, the data are sorted according to that column"""
+
+        # Do not allow sorting if the table is disabled
+        if self._disabled:
+            return
+
+        # Update heading text with icon
+        sort_descending = False
+        for col_id in range(self._num_columns):
+            heading = self._treeview.heading(col_id)['text']
+            if col_id == column_id:
+                # Set the sort icon according to whatever is already there
+                if heading.endswith(self.sort_ascending_icon):
+                    # Change from ascending icon to descending
+                    heading_descending = f'{heading[:-len(self.sort_descending_icon)]}{self.sort_descending_icon}'
+                    self._treeview.heading(col_id, text=heading_descending)
+                    sort_descending = True
+                elif heading.endswith(self.sort_descending_icon):
+                    # Change from descending icon to ascending
+                    heading_ascending = f'{heading[:-len(self.sort_descending_icon)]}{self.sort_ascending_icon}'
+                    self._treeview.heading(col_id, text=heading_ascending)
+                else:
+                    # No current icon - set to descending by default
+                    heading_ascending = f'{heading}{self.sort_ascending_icon}'
+                    self._treeview.heading(col_id, text=heading_ascending)
+            else:
+                # Clear the icon from the heading text if it exists
+                if heading.endswith(self.sort_ascending_icon) or heading.endswith(self.sort_descending_icon):
+                    self._treeview.heading(col_id, text=heading[:-len(self.sort_descending_icon)])
+
+        # Sort the data
+        self.data = sorted(self.data, key=lambda l: l[column_id], reverse=sort_descending)
+
+    def _clear_sort_icons(self):
+        """Clear any icons that have been appended to column headings"""
+        for col_id in range(self._num_columns):
+            heading = self._treeview.heading(col_id)['text']
+            if heading.endswith(self.sort_ascending_icon) or heading.endswith(self.sort_descending_icon):
+                self._treeview.heading(col_id, text=heading[:-len(self.sort_descending_icon)])
+
+    @property
+    def data(self):
+        """Returns all data in the table as a list of lists"""
+        return [self._treeview.item(line)['values'] for line in self._treeview.get_children()]
+
+    @data.setter
+    def data(self, values):
+        """Replaces all data in the table with the given values, which must be a list of lists"""
+        if not all(type(row) in (list, tuple) for row in values):
+            raise ValueError('Table data must be a list of lists')
+        if not all(len(row) == self._num_columns for row in values):
+            raise ValueError('Could not set table data - the number of columns of the table does not match')
+
+        self.clear()
+        for line in values:
+            self._treeview.insert('', 'end', values=line)
+
+    @property
+    def multiple_selection(self):
+        """Allows multiple rows to be selected with shift-click or ctrl-click"""
+        return str(self._treeview.cget('selectmode')) == 'extended'
+
+    @multiple_selection.setter
+    def multiple_selection(self, multiple):
+        """Changes the selection mode between single and multiple"""
+        mode = 'extended' if multiple else 'browse'
+        self._treeview.config(selectmode=mode)
+        # Clear the selection if single selection is enabled
+        if not multiple:
+            self.select_none()
+
+    @property
+    def selected(self):
+        """Returns the selected row, or a list of rows, or None"""
+        selected_ids = self._treeview.selection()
+        if not selected_ids:
+            return None
+        if self.multiple_selection:
+            return [self._treeview.item(row_id)['values'] for row_id in reversed(selected_ids)]
+        else:
+            return self._treeview.item(selected_ids)['values']
+
+    def add_row_at(self, index, data):
+        """Adds a row of data to the table at a given index"""
+        # Check if location is an integer
+        if type(index) != int and index != 'end':
+            raise TypeError(f'index must be an integer. The value provided was {index}')
+        if not type(data) in (list, tuple):
+            raise TypeError(f'row data must be a list')
+        # Check if the number of columns in the data is correct
+        if len(data) != self._num_columns:
+            raise ValueError(f'The number of data arguments given ({len(data)}) does not match '
+                             f'the number of columns in the table ({self._num_columns})')
+
+        self._treeview.insert('', index, values=data)
+
+        # Clear any sort icons if new data is added
+        self._clear_sort_icons()
+
+    def add_row(self, data):
         """Adds a row of data to the table"""
-        pass
+        self.add_row_at('end', data)
+
+    def add_row_to_top(self, data):
+        """Adds a row of data to the start of the table"""
+        self.add_row_at(0, data)
+
+    def clear(self):
+        """Removes all data rows from the table"""
+        for row_id in self._treeview.get_children():
+            self._treeview.delete(row_id)
+
+    def remove_row(self, index):
+        """Removes the specified row from the table, indexed from 0"""
+        row_ids = self._treeview.get_children()
+        if type(index) != int:
+            raise TypeError(f'index must be an integer. The value provided was {index}')
+        if index < 0 or index > len(row_ids) - 1:
+            raise IndexError(f'The index must be between 0 and {len(row_ids) - 1}. '
+                             f'The value of index was {index}')
+        row_data = self._treeview.item(row_ids[index])['values']
+        self._treeview.delete(row_ids[index])
+        return row_data
+
+    def remove_selected(self):
+        """Removes the currently selected row from the table"""
+        row_data = self.selected
+        self._treeview.delete(*self._treeview.selection())
+        return row_data
+
+    def set_column_width(self, column, width):
+        """Sets the width in pixels of the specified column, indexed from 0"""
+        if type(column) != int or column < 0:
+            raise TypeError(f'Column number must be a positive integer. The value given was {column}')
+        if type(width) != int or width <= 0:
+            raise TypeError(f'Column width must be a positive integer. The value given was {width}')
+        self._treeview.column(column, width=width)
+
+    def set_column_widths(self, *widths):
+        """Sets the width in pixels of all columns of the table"""
+        if len(widths) != self._num_columns:
+            raise ValueError(f'The number of arguments supplied ({len(widths)}) does not match '
+                             f'the number of columns in the table ({self._num_columns})')
+        for column, width in enumerate(widths):
+            self.set_column_width(column, width)
+
+    def set_column_alignment(self, column, align):
+        """Sets the alignment of the specified column, indexed from 0"""
+        alignment_mapping = {'left': 'w', 'center': 'center', 'right': 'e'}
+
+        if type(column) != int or column < 0:
+            raise TypeError(f'Column number must be a positive integer. The value given was {column}')
+        if align not in alignment_mapping.keys():
+            raise ValueError(f'Column alignment value must be either "left", "right" or "center". '
+                             f'The value provided was "{align}"')
+        self._treeview.column(column, anchor=alignment_mapping[align])
+
+    def set_column_alignments(self, *aligns):
+        """Sets the text alignment (left, center or right) """
+        if len(aligns) != self._num_columns:
+            raise ValueError(f'The number of arguments supplied ({len(aligns)}) does not match '
+                             f'the number of columns in the table ({self._num_columns})')
+        for column, align in enumerate(aligns):
+            self.set_column_alignment(column, align)
+
+    def select_row(self, index):
+        """Selects a given row in the table, indexed from 0"""
+        all_rows = self._treeview.get_children()
+        if len(self._treeview.get_children()) == 0:
+            raise ValueError(f'Table has no rows to select')
+        if index not in range(len(all_rows)):
+            raise IndexError(f'The index must be between 0 and {len(all_rows) - 1}. '
+                             f'The value of the index specified was {index}.')
+        row_id = all_rows[index]
+        self._treeview.selection_set(row_id)
+        self._treeview.see(row_id)  # Show the selected row (in case it is not be in view)
+
+    def select_all(self):
+        """Selects all rows of the table if multiple selection is enabled.
+        Has no effect if multiple selection is not enabled"""
+        if self.multiple_selection:
+            self._treeview.selection_set(*self._treeview.get_children())
+
+    def select_none(self):
+        """Clears any selected rows in the table"""
+        self._treeview.selection_remove(*self._treeview.selection())
